import 'dart:async';

import 'package:http/http.dart' as http;
import 'package:http/http.dart';
import 'package:meta/meta.dart';
import 'package:osam_http_client/src/exceptions.dart';

String bearer(String token) => 'bearer $token';

abstract class HttpClient {
  const factory HttpClient({@required String host, Duration timeout, bool log}) = _SimpleHttpClient;

  Future<Response> get({@required String path, Map<String, String> headers});

  Future<Response> post({@required String path, Map<String, String> headers, Object body});
}

class _SimpleHttpClient implements HttpClient {
  final Duration timeout;
  final String host;
  final bool log;

  const _SimpleHttpClient(
      {this.host = '', this.timeout = const Duration(seconds: 5), this.log = false});

  Response _checkedResponse(Response response) {
    final statusCode = response.statusCode;
    if (statusCode >= 200 && statusCode < 300) return response;
    if (statusCode == 401) throw HttpUnauthorizedException();
    if (statusCode >= 400 && statusCode < 500) throw HttpClientException();
    if (statusCode >= 500 && statusCode < 600) throw HttpServerException();
    throw HttpUnknownException();
  }

  @override
  Future<Response> get({@required String path, Map<String, String> headers}) {
    if (log) print('path: $path,\nheaders: $headers');
    return http
        .get(Uri.encodeFull('$host$path'), headers: headers)
        .timeout(timeout)
        .catchError((Object error) {
      if (log) print(error);
      throw NoConnectionException();
    }).then((response) {
      if (log) print(response.log());
      return _checkedResponse(response);
    });
  }

  @override
  Future<Response> post({@required String path, Map<String, String> headers, Object body}) {
    if (log) print('path: $path,\nheaders: $headers, \nbody');
    return http
        .post(Uri.encodeFull('$host$path'), headers: headers, body: body)
        .timeout(timeout)
        .catchError((Object error) {
      if (log) print(error);
      throw NoConnectionException();
    }).then((response) {
      if (log) print(response.log());
      return _checkedResponse(response);
    });
  }
}

extension Log on Response {
  String log() {
    return 'headers: $headers\nstatusCode: $statusCode\nbody: $body';
  }
}
